import numpy as np

from typing import Generator, Literal
import math


def batch_slices(
    size: int,
    batch_size: int,
    drop_remainder: bool = False,
) -> Generator[tuple[int, int], None, None]:
    for start in range(0, size, batch_size):
        stop = start + batch_size

        if stop >= size:
            if drop_remainder:
                break

            stop = size

        yield start, stop


def batch_indicies(
    size: int,
    batch_size: int,
    drop_remainder: bool = False,
) -> Generator[np.ndarray, None, None]:
    for start, stop in batch_slices(size, batch_size, drop_remainder):
        yield np.arange(start, stop)


def batch_indicies_split(
    size: int,
    batch_size: int,
    split_index: Literal["train", "valid", "test"] = "train",
    valid_part: int = 10,
    kfold_index: int = 0,
    drop_remainder: bool = False,
) -> Generator[np.ndarray, None, None]:
    split_int = {"train": 0, "valid": 1, "test": 2}[split_index]
    cuts = [
        0,
        math.ceil(size * (1 - 2 / valid_part)),
        math.ceil(size * (1 - 1 / valid_part)),
        size,
    ]
    start, stop = cuts[split_int], cuts[split_int + 1]
    kfold_size = size // valid_part

    for indicies in batch_indicies(stop - start, batch_size, drop_remainder):
        indicies = indicies + start
        indicies = indicies + kfold_size * kfold_index

        indicies = indicies % size
        yield indicies



import math
import os
import tempfile
from typing import Self

import numpy as np
from numpy.typing import DTypeLike


__all__ = ["Vector"]


class Vector:
    _mmap: None | np.memmap

    def __init__(self: Self) -> None:
        self._mmap = None
        self._len = 0

    @property
    def array(self: Self) -> np.ndarray:
        if self._mmap is None:
            return np.empty((0, 0), dtype=np.float32)

        self._mmap.flush()
        return self._mmap[: self._len]

    def __len__(self: Self) -> int:
        return self._len if self._mmap is not None else 0

    def _init_mmap(self: Self, shape: tuple[int, ...], dtype: DTypeLike) -> None:
        self.file = tempfile.NamedTemporaryFile(dir=".")
        self._mmap = np.memmap(
            self.file,
            dtype,
            "w+",
            shape=(1, *shape),
        )

    def extend(self: Self, value: np.ndarray) -> None:
        if self._mmap is None:
            self._init_mmap(value.shape[1:], value.dtype)

        assert self._mmap is not None, "Memory-mapped array is not initialized"

        if value.dtype != self._mmap.dtype:
            error = f"Got dtype {value.dtype} != {self._mmap.dtype}."
            raise ValueError(error)

        if value.shape[1:] != self._mmap.shape[1:]:
            error = f"Got shape {value.shape[1:]} != {self._mmap.shape[1:]}."
            raise ValueError(error)

        length = self._len + value.shape[0]
        length = 2 ** math.ceil(math.log2(length))

        if length > self._mmap.shape[0]:
            file = tempfile.NamedTemporaryFile(dir=".")
            mmap = np.memmap(
                file,
                self._mmap.dtype,
                "w+",
                shape=(length, *self._mmap.shape[1:]),
            )
            mmap[: self._len] = self._mmap[: self._len]
            mmap.flush()

            self.file.close()
            self.file = file
            self._mmap = mmap

        self._mmap[self._len : self._len + value.shape[0]] = value
        self._len += value.shape[0]

    def append(self: Self, value: np.ndarray) -> None:
        self.extend(value[np.newaxis])

    def save(self: Self, path: str) -> None:
        if self._mmap is None:
            raise ValueError("No data to save.")

        self._mmap.flush()

        try:
            mmap = np.memmap(
                path,
                self._mmap.dtype,
                "w+",
                shape=(self._len, *self._mmap.shape[1:]),
            )
            mmap[:] = self._mmap[: self._len]
            mmap.flush()

        # if an error occurs, remove the file
        except Exception as e:
            os.remove(path)
            raise e



from typing import Callable, Self

import numpy as np


class FixedSeed:
    def __init__(self, seed: int | None) -> None:
        self._seed = seed

    def __enter__(self: Self) -> None:
        if self._seed is None:
            return

        self.state = np.random.get_state()
        np.random.seed(self._seed)

    def __exit__(self: Self, *_) -> None:
        if self._seed is None:
            return

        np.random.set_state(self.state)


def shuffle_blocks(
    size: int,
    size_perm: int = 1024,
    seed: int | None = None,
) -> Callable[[np.ndarray], np.ndarray]:
    # permutation for most of the array
    with FixedSeed(seed):
        perm_tot = np.random.uniform(size=size_perm).argsort()

    # threshold for the remaining elements
    c = size // size_perm * size_perm

    # number of remaining elements, at least one to ensure no no null division
    n = max(1, size % size_perm)

    # permutation for the remaining elements
    perm_rem = np.random.uniform(size=n).argsort()

    def apply_perm(x):
        div = x // size_perm
        rem = x % size_perm
        return np.where(
            x < c,
            perm_tot[rem] + div * size_perm,
            perm_rem[rem % n] + div * size_perm,
        )

    return apply_perm


def shuffle_bands(
    size: int,
    size_perm: int = 8,
) -> Callable[[np.ndarray], np.ndarray]:
    # permutation for offsets in modulus equivalance classes
    perm = np.arange(size_perm)

    # has to be multiplied by size_perm to stay in the same class
    perm = perm * size_perm
    # controll the distance between the consecutive elements
    perm = perm * (size // (size_perm - 1))

    # size of the largest multiple of size_perm smaller than size
    size_floor = size // size_perm * size_perm

    def apply_perm(x):
        return np.where(
            x < size_floor,
            (x + perm[x % size_perm]) % size_floor,
            x,
        )

    return apply_perm


def shuffle_fast(
    size: int,
    size_perm_band: int = 8,
    size_perm_block: int = 1024,
    seed: int | None = None,
) -> Callable[[np.ndarray], np.ndarray]:
    shuffle1 = shuffle_blocks(size, size_perm_block, seed)
    shuffle2 = shuffle_bands(size, size_perm_band)

    def apply(x):
        x = shuffle1(x)
        x = (x + size // 2) % size
        x = shuffle2(x)
        return x

    return apply



from beartype.claw import beartype_this_package

beartype_this_package()


from ._vector import Vector
from ._shuffle import shuffle_fast, shuffle_blocks, shuffle_bands
from ._index import batch_slices, batch_indicies, batch_indicies_split

__version__ = "0.1.11"

__all__ = [
    "__version__",
    "Vector",
    "shuffle_fast",
    "shuffle_blocks",
    "shuffle_bands",
    "batch_slices",
    "batch_indicies",
    "batch_indicies_split",
]



from setuptools import setup, find_packages
from memmpy import __version__

with open("README.md", "r") as f:
    long_description = f.read()


with open("requirements.txt") as f:
    requirements = f.read().splitlines()


setup(
    name="memmpy",
    version=__version__,
    packages=find_packages(where="."),
    python_requires=">=3.10",
    install_requires=requirements,
    description="Memory mapped of datasets",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pwolle/memmpy",
    author="Paul Wollenhaupt",
    author_email="paul.wollenhaupt@gmail.com",
)
